/**
 *     __  ____                 _   __                
 *    /  |/  (_)_____________  / | / /___ _   ______ _
 *   / /|_/ / / ___/ ___/ __ \/  |/ / __ \ | / / __ `/
 *  / /  / / / /__/ /  / /_/ / /|  / /_/ / |/ / /_/ / 
 * /_/  /_/_/\___/_/   \____/_/ |_/\____/|___/\__,_/  
 *                                                  
 * NovaVGA Arduino Library
 * v 1.0 (10/20/2015)
 * Copyright (c) 2015 MicroNova, LLC
 * www.micro-nova.com
 * 
 * TETRIS EXAMPLE
 *
 *   The game as-is uses input from the serial port for controls. The wasd keys
 *   can be used like a normal Tetris game - Left/Right ('a' and 'd') to move, Up
 *   ('w') to rotate and Down ('s') to move down faster. Use any terminal program
 *   that doesn't require pressing Enter to send the text, such as Putty.
 * 
 *   Based on: http://www.gamedev.net/topic/192483-tetris-clone-in-an-hour-with-c/
 */

#include "NovaVGAFast.h"

const int TILESIZE   = 8;   // Size of the square Tetris blocks
const int EDGEWIDTH  = 1;   // Width of the left edge of the map, in tiles
const int MAPWIDTH   = 10;  // Width of the Tetris map, in tiles
const int MAPHEIGHT  = 15;  // Height of the Tetris map, in tiles
const int STATSWIDTH = 10;  // Width of the stats/preview section, in tiles
const int TEXTSTART  = (EDGEWIDTH+MAPWIDTH+STATSWIDTH/4)*TILESIZE;

const byte EDGECOLOR     = NovaVGA.Gray;
const byte MAPBACKGND    = NovaVGA.Black;
const byte STATUSBACKGND = NovaVGA.Gray;
const byte TEXTCOLOR     = NovaVGA.Blue;

struct Piece {
  byte color[4][4];
  Point p;
};

byte map_[MAPWIDTH][MAPHEIGHT+1]; // The game map

Piece sPrePiece_;  // Preview piece
Piece sPiece_;     // Current piece

unsigned long startTime_; // Used in timing
bool started_ = false; // Used by NewBlock()
int lines_ = 0;

void setup()
{
  // Setup serial port
  Serial.begin(9600);
  
  // Seed PRNG
  randomSeed(analogRead(0));
  
  // Setup VGA controller
  NovaVGA.init();

  tetrisInit();
}

void loop()
{
  if(Serial.available())
  {
    /* Read the first available character and toss out any others to avoid
     * the game responding to keypresses even after the key was released.
     */
    int ch = Serial.read();
    while(Serial.available()) { Serial.read(); }
    
    if(ch==(int)'w')      // Check for up (w)
    {
      rotateBlock();
    }
    else if(ch==(int)'a') // Check for left (a)
    {
      move(-1,0);
    }
    else if(ch==(int)'s') // Check for down (s)
    {
      move(0,1);
    }
    else if(ch==(int)'d') // Check for move right (d)
    {
      move(1,0);
    }
  }
  
  gameLoop();
}

void tetrisInit()
{
  startTime_ = millis();
  started_ = false;
  lines_ = 0;

  // Set the tetris grid background
  for(int x = 0; x < MAPWIDTH; x++)
  {
    for(int y = 0; y < MAPHEIGHT+1; y++)
    {
      if(y == MAPHEIGHT) //makes Y-collision easier.
        map_[x][y] = NovaVGA.None;
      else
        map_[x][y] = MAPBACKGND;
    }
  }
  
  // Draw the left edge (static)
  NovaVGA.fillRect(0, 0, (EDGEWIDTH)*TILESIZE,
           (MAPHEIGHT)*TILESIZE, EDGECOLOR);

  // Draw the status bar background
  NovaVGA.fillRect((EDGEWIDTH+MAPWIDTH)*TILESIZE, 0, (STATSWIDTH)*TILESIZE,
           (MAPHEIGHT)*TILESIZE, STATUSBACKGND);
  
  NovaVGA.drawString("Next:", TEXTSTART, 4, TEXTCOLOR);
  NovaVGA.drawString("Lines:", TEXTSTART, 80, TEXTCOLOR);
  
  newBlock();
  drawMap();
}

void gameLoop()
{
  if( (millis() - startTime_) > 1000)
  {
    move(0,1);
    startTime_ = millis();
  }
}

// x and y are in blocks not pixels
void drawTile(int x,int y, byte color)
{
  NovaVGA.fillRect(x*TILESIZE, y*TILESIZE, TILESIZE, TILESIZE, color);
}

// Update tetris grid and scores
void drawMap()
{
  int x, y;
  
  // Draw the map
  for(x = 0; x < MAPWIDTH; x++)
    for(y = 0; y < MAPHEIGHT; y++)
      drawTile(x+EDGEWIDTH, y, map_[x][y]);

  // Draw moving block
  for(x = 0; x < 4; x++)
    for(y = 0; y < 4; y++)
      if(sPiece_.color[x][y] != NovaVGA.None)
        drawTile(sPiece_.p.x+x, sPiece_.p.y+y, sPiece_.color[x][y]);

  NovaVGA.fillRect(TEXTSTART, 92, NovaVGA.SCREEN_WIDTH - TEXTSTART, NovaVGA.CHAR_HEIGHT, STATUSBACKGND);
  NovaVGA.drawString(String(lines_), (EDGEWIDTH+MAPWIDTH+STATSWIDTH/4)*TILESIZE, 92, TEXTCOLOR);
}

void newBlock()
{
  int i,j,x,y;

  // Initialize the piece to all blank.
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      sPiece_.color[i][j] = NovaVGA.None;

  sPiece_.p.x=EDGEWIDTH+MAPWIDTH/2-2;
  sPiece_.p.y=-1;

  // Check if the game has started_ yet
  if(started_ == false)
  {
    // Generate a piece right off.
    // From now on, use previous preview block.
    started_=true;
    
    switch (random(7))
    {
      case 0: // I
        sPiece_.color[1][0] = NovaVGA.Cyan;
        sPiece_.color[1][1] = NovaVGA.Cyan;
        sPiece_.color[1][2] = NovaVGA.Cyan;
        sPiece_.color[1][3] = NovaVGA.Cyan;
        sPiece_.p.y=0;  
        break;
      case 1: // O
        sPiece_.color[1][1] = NovaVGA.Yellow;
        sPiece_.color[1][2] = NovaVGA.Yellow;
        sPiece_.color[2][1] = NovaVGA.Yellow;
        sPiece_.color[2][2] = NovaVGA.Yellow;
        break;
      case 2: // T
        sPiece_.color[1][1] = NovaVGA.Magenta;
        sPiece_.color[0][2] = NovaVGA.Magenta;
        sPiece_.color[1][2] = NovaVGA.Magenta;
        sPiece_.color[2][2] = NovaVGA.Magenta;
        break;
      case 3: // Z
        sPiece_.color[0][1] = NovaVGA.Red;
        sPiece_.color[1][1] = NovaVGA.Red;
        sPiece_.color[1][2] = NovaVGA.Red;
        sPiece_.color[2][2] = NovaVGA.Red;
        break;
      case 4: // S
        sPiece_.color[2][1] = NovaVGA.Green;
        sPiece_.color[1][1] = NovaVGA.Green;
        sPiece_.color[1][2] = NovaVGA.Green;
        sPiece_.color[0][2] = NovaVGA.Green;
        break;
      case 5: // L
        sPiece_.color[1][1] = NovaVGA.Orange;
        sPiece_.color[2][1] = NovaVGA.Orange;
        sPiece_.color[2][2] = NovaVGA.Orange;
        sPiece_.color[2][3] = NovaVGA.Orange;
        break;
      case 6: // J
        sPiece_.color[2][1] = NovaVGA.Blue;
        sPiece_.color[1][1] = NovaVGA.Blue;
        sPiece_.color[1][2] = NovaVGA.Blue;
        sPiece_.color[1][3] = NovaVGA.Blue;
        break;
    }
  }
  else
  {
    for(i = 0; i < 4; i++)
      for(j = 0; j < 4; j++)
        sPiece_.color[i][j] = sPrePiece_.color[i][j];
  }

  // Pick new block and set as preview
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      sPrePiece_.color[i][j] = NovaVGA.None;

  sPrePiece_.p.x = EDGEWIDTH+MAPWIDTH+STATSWIDTH/2-2;
  sPrePiece_.p.y = STATSWIDTH/4;

  switch (random(7))
  {
    case 0: // I
      sPrePiece_.color[1][0] = NovaVGA.Cyan;
      sPrePiece_.color[1][1] = NovaVGA.Cyan;
      sPrePiece_.color[1][2] = NovaVGA.Cyan;
      sPrePiece_.color[1][3] = NovaVGA.Cyan;
      break;
    case 1: // O
      sPrePiece_.color[1][1] = NovaVGA.Yellow;
      sPrePiece_.color[1][2] = NovaVGA.Yellow;
      sPrePiece_.color[2][1] = NovaVGA.Yellow;
      sPrePiece_.color[2][2] = NovaVGA.Yellow;
      break;
    case 2: // T
      sPrePiece_.color[1][1] = NovaVGA.Magenta;
      sPrePiece_.color[0][2] = NovaVGA.Magenta;
      sPrePiece_.color[1][2] = NovaVGA.Magenta;
      sPrePiece_.color[2][2] = NovaVGA.Magenta;
      break;
    case 3:// Z
      sPrePiece_.color[0][1] = NovaVGA.Red;
      sPrePiece_.color[1][1] = NovaVGA.Red;
      sPrePiece_.color[1][2] = NovaVGA.Red;
      sPrePiece_.color[2][2] = NovaVGA.Red;
      break;
    case 4:// S
      sPrePiece_.color[2][1] = NovaVGA.Green;
      sPrePiece_.color[1][1] = NovaVGA.Green;
      sPrePiece_.color[1][2] = NovaVGA.Green;
      sPrePiece_.color[0][2] = NovaVGA.Green;
      break;
    case 5:// L
      sPrePiece_.color[1][1] = NovaVGA.Orange;
      sPrePiece_.color[2][1] = NovaVGA.Orange;
      sPrePiece_.color[2][2] = NovaVGA.Orange;
      sPrePiece_.color[2][3] = NovaVGA.Orange;
      break;
    case 6:// J
      sPrePiece_.color[2][1] = NovaVGA.Blue;
      sPrePiece_.color[1][1] = NovaVGA.Blue;
      sPrePiece_.color[1][2] = NovaVGA.Blue;
      sPrePiece_.color[1][3] = NovaVGA.Blue;
      break;
  }

  // Draw preview block
  for(x = 0; x < 4; x++)
    for(y = 0; y < 4; y++)
      if(sPrePiece_.color[x][y] != NovaVGA.None)
        drawTile(sPrePiece_.p.x+x, sPrePiece_.p.y+y, sPrePiece_.color[x][y]);
      else
        NovaVGA.fillRect((sPrePiece_.p.x+x)*TILESIZE, (sPrePiece_.p.y+y)*TILESIZE,
                 TILESIZE, TILESIZE, STATUSBACKGND);

  drawMap();
}

void rotateBlock()
{
  int i, j;
  byte temp[4][4];

  // Copy and rotate the piece to a temporary array
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      temp[3-j][i] = sPiece_.color[i][j];

  // Check collision of the temporary array with map borders
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      if(temp[i][j] != NovaVGA.None)
        if(sPiece_.p.x + i < EDGEWIDTH ||
          sPiece_.p.x + i > MAPWIDTH + EDGEWIDTH - 1 ||
          sPiece_.p.y + j < 0 || sPiece_.p.y + j > MAPHEIGHT - 1)
          return;

  // Check collision of the temporary array with the blocks on the map
  for(int x = 0; x < MAPWIDTH; x++)
    for(int y = 0; y < MAPHEIGHT; y++)
      if(x >= sPiece_.p.x && x < sPiece_.p.x + 4)
        if(y >= sPiece_.p.y && y < sPiece_.p.y + 4)
          if(map_[x][y] != MAPBACKGND)
            if(temp[x - sPiece_.p.x][y - sPiece_.p.y] != NovaVGA.None)
              return;

  // Copy the rotated temporary array to the original piece
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      sPiece_.color[i][j] = temp[i][j];
  
  drawMap();

  return;
}

void move(int x, int y)
{
  if(collisionTest(x, y))
  {
    if(y == 1)
    {
      if(sPiece_.p.y < 1)
      {
        // You lose! New game.
        tetrisInit();
      }
      else
      {
        int i,j;
        
        // New block time! Add this one to the list!
        for(i = 0; i < 4; i++)
          for(j = 0; j < 4; j++)
            if(sPiece_.color[i][j] != NovaVGA.None)
              map_[sPiece_.p.x-EDGEWIDTH+i][sPiece_.p.y+j] = sPiece_.color[i][j];

        // Check for cleared row!
        for(j = 0; j < MAPHEIGHT; j++)
        {
          bool filled=true;
          for(i = 0; i < MAPWIDTH; i++)
            if(map_[i][j] == MAPBACKGND)
              filled = false;

          if(filled)
            removeRow(j);
        }
        newBlock();
      }
    }
  }
  else
  {
    sPiece_.p.x += x;
    sPiece_.p.y += y;
  }

  drawMap();
}

bool collisionTest(int nx, int ny)
{
  int newx = sPiece_.p.x + nx - EDGEWIDTH;
  int newy = sPiece_.p.y + ny;

  int i, j, x, y;

  //check collision of the temporary array with map borders
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      if(sPiece_.color[i][j] != NovaVGA.None)
        if(newx + i < 0 || newx + i > MAPWIDTH - 1 ||
          newy + j < 0 || newy + j > MAPHEIGHT - 1)
          return true;

  //check collision of the piece with the blocks on the map
  for(x = 0; x < MAPWIDTH; x++)
    for(y = 0; y < MAPHEIGHT; y++)
      if(x >= newx && x < newx + 4)
        if(y >= newy && y < newy + 4)
          if(map_[x][y] != MAPBACKGND)
            if(sPiece_.color[x - newx][y - newy] != NovaVGA.None)
              return true;
  return false;
}

void removeRow(int row)
{
  int x,y;

  for(x = 0; x < MAPWIDTH; x++)
    for(y = row; y > 0; y--)
      map_[x][y] = map_[x][y-1];

  lines_++;
}
